@@ -6,8 +6,8 @@ from .utils import as_float
 from .event import Event, EventType
 import re
 
-class Individual(object):
 
+class Individual(object):
     def __init__(self, id, susceptibility, model, logger):
         self.id = id
         self.model = model
@@ -26,24 +26,19 @@ class Individual(object):
 
     @property
     def group(self):
-        return self.id.rsplit('_', 1)[0] if '_' in self.id else ''
+        return self.id.rsplit("_", 1)[0] if "_" in self.id else ""
 
     def __str__(self):
         return self.id
 
     def quarantine(self, **kwargs):
-        if 'till' in kwargs:
-            till = kwargs['till']
+        if "till" in kwargs:
+            till = kwargs["till"]
         else:
-            raise ValueError(
-                'No till parameter is specified for quarantine event.')
+            raise ValueError("No till parameter is specified for quarantine event.")
         self.quarantined = till
         return [
-            Event(
-                till,
-                EventType.REINTEGRATION,
-                target=self.id,
-                logger=self.logger)
+            Event(till, EventType.REINTEGRATION, target=self.id, logger=self.logger)
         ]
 
     def reintegrate(self):
@@ -53,32 +48,44 @@ class Individual(object):
     def symptomatic_infect(self, time, **kwargs):
         self.r0 = self.model.draw_random_r0(symptomatic=True, group=self.group)
         self.incubation_period = self.model.draw_random_incubation_period(
-            group=self.group)
+            group=self.group
+        )
 
-        by = kwargs.get('by', None)
+        #
+        # infect others
+        (
+            x_grid,
+            trans_prob,
+            infect_params,
+        ) = self.model.get_symptomatic_transmission_probability(
+            self.incubation_period, self.r0
+        )
 
-        if 'leadtime' in kwargs and kwargs['leadtime'] is not None:
+        by = kwargs.get("by", None)
+
+        if "leadtime" in kwargs and kwargs["leadtime"] is not None:
             if by is not None:
                 raise ValueError(
-                    'leadtime is only allowed during initialization of infection event (no by option.)'
+                    "leadtime is only allowed during initialization of infection event (no by option.)"
                 )
-            if kwargs['leadtime'] == 'any':
-                lead_time = np.random.uniform(
-                    0, self.incubation_period * 2 / 3 + 8)
-            elif kwargs['leadtime'] == 'asymptomatic':
+            if kwargs["leadtime"] == "any":
+                lead_time = np.random.uniform(0, x_grid[-1])
+            elif kwargs["leadtime"] == "asymptomatic":
                 lead_time = np.random.uniform(0, self.incubation_period)
             else:
-                lead_time = as_float(kwargs['leadtime'], "--leadtime can only be any, asymptomatic, or a fixed number")
+                lead_time = as_float(
+                    kwargs["leadtime"],
+                    "--leadtime can only be any, asymptomatic, or a fixed number",
+                )
         else:
             lead_time = 0
 
         self.infected = float(time - lead_time)
 
-        if 'handle_symptomatic' not in kwargs or kwargs[
-                'handle_symptomatic'] is None:
-            handle_symptomatic = ['remove', 1]
+        if "handle_symptomatic" not in kwargs or kwargs["handle_symptomatic"] is None:
+            handle_symptomatic = ["remove", 1]
         else:
-            handle_symptomatic = kwargs['handle_symptomatic']
+            handle_symptomatic = kwargs["handle_symptomatic"]
         # show symptom
         kept = True
         evts = []
@@ -92,25 +99,32 @@ class Individual(object):
                     symp_time,
                     EventType.SHOW_SYMPTOM,
                     target=self.id,
-                    logger=self.logger))
+                    logger=self.logger,
+                )
+            )
         #
         if self.quarantined and self.quarantined > symp_time:
             # show symptom during quarantine, ok
             pass
-        elif handle_symptomatic[0] in ('remove', 'keep'):
+        elif handle_symptomatic[0] in ("remove", "keep"):
             if len(handle_symptomatic) == 1:
                 proportion = 1
             else:
-                proportion = as_float(handle_symptomatic[1], "Proportion in --handle-symptomatic remove/keep prop should be a float number")
+                proportion = as_float(
+                    handle_symptomatic[1],
+                    "Proportion in --handle-symptomatic remove/keep prop should be a float number",
+                )
             if proportion > 1 or proportion < 0:
                 raise ValueError(
                     f'Proportion in "--handle-symptomatic remove/keep prop" should be a float number between 0 and 1: {proportion} provided'
                 )
-            if (handle_symptomatic[0] == 'keep' and
-                    np.random.uniform(0, 1, 1)[0] > proportion) or (
-                        handle_symptomatic[0] == 'remove' and
-                        (proportion == 1 or
-                         np.random.uniform(0, 1, 1)[0] <= proportion)):
+            if (
+                handle_symptomatic[0] == "keep"
+                and np.random.uniform(0, 1, 1)[0] > proportion
+            ) or (
+                handle_symptomatic[0] == "remove"
+                and (proportion == 1 or np.random.uniform(0, 1, 1)[0] <= proportion)
+            ):
                 kept = False
                 if symp_time >= 0:
                     evts.append(
@@ -119,23 +133,30 @@ class Individual(object):
                             symp_time,
                             EventType.REMOVAL,
                             target=self.id,
-                            logger=self.logger))
+                            logger=self.logger,
+                        )
+                    )
                 else:
                     # just remove
                     self.logger.write(
                         f'{self.logger.id}\t{time:.2f}\t{EventType.WARNING.name}\t{self.id}\tmsg="Individual not removed before it show symptom before {time}"\n'
                     )
 
-        elif handle_symptomatic[0].startswith('quarantine'):
-            if handle_symptomatic[0] == 'quarantine':
+        elif handle_symptomatic[0].startswith("quarantine"):
+            if handle_symptomatic[0] == "quarantine":
                 quarantine_duration = 14
             else:
-                quarantine_duration = as_float(handle_symptomatic[0].split(
-                        '_', 1)[1], "quanrantine duration should be specified as quarantine_DURATION")
+                quarantine_duration = as_float(
+                    handle_symptomatic[0].split("_", 1)[1],
+                    "quanrantine duration should be specified as quarantine_DURATION",
+                )
             if len(handle_symptomatic) == 1:
                 proportion = 1
             else:
-                proportion = as_float(handle_symptomatic[1], "Proportion in --handle-symptomatic quarantine_DURATION prop should be a float number")
+                proportion = as_float(
+                    handle_symptomatic[1],
+                    "Proportion in --handle-symptomatic quarantine_DURATION prop should be a float number",
+                )
             if proportion > 1 or proportion < 0:
                 raise ValueError(
                     f'Proportion in "--handle-symptomatic quarantine_DUURATION prop" should be a float number between 0 and 1: {proportion} provided'
@@ -150,33 +171,32 @@ class Individual(object):
                             EventType.QUARANTINE,
                             target=self.id,
                             logger=self.logger,
-                            till=symp_time + quarantine_duration))
+                            till=symp_time + quarantine_duration,
+                        )
+                    )
                 else:
                     self.quarantined = symp_time + quarantine_duration
         else:
             raise ValueError(
                 f'Unrecognizable symptomatic case handling method: {" ".join(handle_symptomatic)}'
             )
-        #
-        # infect others
-        x_grid, trans_prob = self.model.get_symptomatic_transmission_probability(
-            self.incubation_period, self.r0)
+
+        self.infect_params = infect_params
         # infect only before removal or quarantine
         if kept:
             x_before = x_grid
         else:
             # remove or quanratine
-            x_before = [
-                x for x in x_grid if x < self.incubation_period - lead_time
-            ]
-        infected = np.random.binomial(1, trans_prob[:len(x_before)],
-                                      len(x_before))
+            x_before = [x for x in x_grid if x < self.incubation_period - lead_time]
+        infected = np.random.binomial(1, trans_prob[: len(x_before)], len(x_before))
         presymptomatic_infected = [
-            xx for xx, ii in zip(x_before, infected)
+            xx
+            for xx, ii in zip(x_before, infected)
             if ii and xx < self.incubation_period
         ]
         symptomatic_infected = [
-            xx for xx, ii in zip(x_before, infected)
+            xx
+            for xx, ii in zip(x_before, infected)
             if ii and xx >= self.incubation_period
         ]
         if self.quarantined:
@@ -188,7 +208,9 @@ class Individual(object):
                             EventType.INFECTION_AVOIDED,
                             target=self.id,
                             logger=self.logger,
-                            by=self.id))
+                            by=self.id,
+                        )
+                    )
                     infected[idx] = 0
         #
         for x, infe in zip(x_before, infected):
@@ -200,29 +222,34 @@ class Individual(object):
                         target=None,
                         logger=self.logger,
                         by=self.id,
-                        handle_symptomatic=kwargs.get('handle_symptomatic',
-                                                      None)))
+                        handle_symptomatic=kwargs.get("handle_symptomatic", None),
+                    )
+                )
 
         evts.append(
             Event(
                 time + x_grid[-1] - lead_time,
                 EventType.RECOVER,
                 target=self.id,
-                logger=self.logger))
+                logger=self.logger,
+            )
+        )
         if by:
-            params = [f'by={by}']
+            params = [f"by={by}"]
         elif lead_time:
-            params = [f'leadtime={lead_time:.2f}']
+            params = [f"leadtime={lead_time:.2f}"]
         else:
             params = []
         #
-        params.extend([
-            f'r0={self.r0:.2f}',
-            f'r={sum(infected)}',
-            f'r_presym={len(presymptomatic_infected)}',
-            f'r_sym={len(symptomatic_infected)}',
-            f'incu={self.incubation_period:.2f}',
-        ])
+        params.extend(
+            [
+                f"r0={self.r0:.2f}",
+                f"r={sum(infected)}",
+                f"r_presym={len(presymptomatic_infected)}",
+                f"r_sym={len(symptomatic_infected)}",
+                f"incu={self.incubation_period:.2f}",
+            ]
+        )
         self.logger.write(
             f'{self.logger.id}\t{time:.2f}\t{EventType.INFECTION.name}\t{self.id}\t{",".join(params)}\n'
         )
@@ -232,20 +259,29 @@ class Individual(object):
         self.r0 = self.model.draw_random_r0(symptomatic=False)
         self.incubation_period = -1
 
-        by = kwargs.get('by',)
+        by = kwargs.get("by")
 
-        if 'leadtime' in kwargs and kwargs['leadtime'] is not None:
+        (
+            x_grid,
+            trans_prob,
+            infect_params,
+        ) = self.model.get_asymptomatic_transmissibility_probability(self.r0)
+
+        if "leadtime" in kwargs and kwargs["leadtime"] is not None:
             if by is not None:
                 raise ValueError(
-                    'leadtime is only allowed during initialization of infection event (no by option.)'
+                    "leadtime is only allowed during initialization of infection event (no by option.)"
                 )
 
-            if kwargs['leadtime'] in ('any', 'asymptomatic'):
+            if kwargs["leadtime"] in ("any", "asymptomatic"):
                 # this is the first infection, the guy should be asymptomatic, but
                 # could be anywhere in his incubation period
-                lead_time = np.random.uniform(0, 10)
+                lead_time = np.random.uniform(0, x_grid[-1])
             else:
-                lead_time = as_float(kwargs['leadtime'], "--leadtime can only be any, asymptomatic, or a fixed number")
+                lead_time = min(as_float(
+                    kwargs["leadtime"],
+                    "--leadtime can only be any, asymptomatic, or a fixed number",
+                ),  x_grid[-1])
         else:
             lead_time = 0
 
@@ -254,11 +290,12 @@ class Individual(object):
         # REMOVAL ...
         evts = []
         #
-        x_grid, trans_prob = self.model.get_asymptomatic_transmission_probability(
-            self.r0)
+        self.infect_params = infect_params
 
         if lead_time > 0:
             idx = int(lead_time / self.model.params.simulation_interval)
+            if idx >= len(trans_prob):
+                idx -= 1
             trans_prob = trans_prob[idx:]
             x_grid = x_grid[idx:]
             x_grid = x_grid - x_grid[0]
@@ -275,7 +312,9 @@ class Individual(object):
                             EventType.INFECTION_AVOIDED,
                             target=self.id,
                             logger=self.logger,
-                            by=self.id))
+                            by=self.id,
+                        )
+                    )
                     infected[idx] = 0
         #
         for x, infe in zip(x_grid, infected):
@@ -287,42 +326,97 @@ class Individual(object):
                         target=None,
                         logger=self.logger,
                         by=self.id,
-                        handle_symptomatic=kwargs.get('handle_symptomatic',
-                                                      None)))
+                        handle_symptomatic=kwargs.get("handle_symptomatic", None),
+                    )
+                )
         evts.append(
             Event(
-                time + x_grid[-1],
-                EventType.RECOVER,
-                target=self.id,
-                logger=self.logger))
+                time + x_grid[-1], EventType.RECOVER, target=self.id, logger=self.logger
+            )
+        )
         if by:
-            params = [f'by={by}']
+            params = [f"by={by}"]
         elif lead_time > 0:
-            params = [f'leadtime={lead_time:.2f}']
+            params = [f"leadtime={lead_time:.2f}"]
         else:
             params = []
         #
-        params.extend([
-            f'r0={self.r0:.2f}', f'r={asymptomatic_infected}',
-            f'r_asym={asymptomatic_infected}'
-        ])
+        params.extend(
+            [
+                f"r0={self.r0:.2f}",
+                f"r={asymptomatic_infected}",
+                f"r_asym={asymptomatic_infected}",
+            ]
+        )
         self.logger.write(
             f'{self.logger.id}\t{time:.2f}\t{EventType.INFECTION.name}\t{self.id}\t{",".join(params)}\n'
         )
         return evts
 
+    def transmissibility(self, time):
+        # return transmissibility at specified time
+        interval = time - self.infected
+        if not self.infect_params:
+            raise ValueError(
+                "This transmissibility model does not record infection parameters."
+            )
+
+        infect_time, peak_time, duration, peak_transmissibility = self.infect_params
+        if interval < infect_time:
+            return 0
+        elif interval < peak_time:
+            return (
+                peak_transmissibility
+                * (interval - infect_time)
+                / (peak_time - infect_time)
+            )
+        else:
+            # assuming that viral load will decrease a lot slower than transmissibility
+            # we use 2 * duration to simulate this effect
+            return max(
+                0,
+                peak_transmissibility * (duration - interval) / (duration - peak_time),
+            )
+
+    def viral_load(self, time):
+        # return transmissibility at specified time
+        interval = time - self.infected
+        if not self.infect_params:
+            raise ValueError(
+                "This transmissibility model does not record infection parameters."
+            )
+
+        infect_time, peak_time, duration, peak_transmissibility = self.infect_params
+        if interval < infect_time:
+            return 0
+        elif interval < peak_time:
+            return (
+                peak_transmissibility
+                * (interval - infect_time)
+                / (peak_time - infect_time)
+            )
+        else:
+            # assuming that viral load will decrease a lot slower than transmissibility
+            # we use 2 * duration to simulate this effect
+            return max(
+                0,
+                peak_transmissibility
+                * (2 * duration - interval)
+                / (2 * duration - peak_time),
+            )
+
     def infect(self, time, **kwargs):
         if isinstance(self.infected, float):
-            by_id = '.' if kwargs['by'] is None else kwargs['by']
+            by_id = "." if kwargs["by"] is None else kwargs["by"]
             self.logger.write(
-                f'{self.logger.id}\t{time:.2f}\t{EventType.INFECTION_IGNORED.name}\t{self.id}\tby={by_id}\n'
+                f"{self.logger.id}\t{time:.2f}\t{EventType.INFECTION_IGNORED.name}\t{self.id}\tby={by_id}\n"
             )
             return []
 
         if self.susceptibility < 1 and rand() > self.susceptibility:
-            by_id = '.' if kwargs['by'] is None else kwargs['by']
+            by_id = "." if kwargs["by"] is None else kwargs["by"]
             self.logger.write(
-                f'{self.logger.id}\t{time:.2f}\t{EventType.INFECTION_FAILED.name}\t{self.id}\tby={by_id},reson=susceptibility\n'
+                f"{self.logger.id}\t{time:.2f}\t{EventType.INFECTION_FAILED.name}\t{self.id}\tby={by_id},reson=susceptibility\n"
             )
             return []
 
@@ -333,41 +427,44 @@ class Individual(object):
 
 
 class Population(object):
-
     def __init__(self, popsize, model, vicinity, logger):
         self.individuals = {}
-        self.group_sizes = {(ps.split('=', 1)[0] if '=' in ps else ''):
-                        0 for ps in popsize}
-        self.max_ids = {x:y for x,y in self.group_sizes.items() }
-        self.subpop_from_id = re.compile('^(.*?)[\d]+$')
+        self.group_sizes = {
+            (ps.split("=", 1)[0] if "=" in ps else ""): 0 for ps in popsize
+        }
+        self.max_ids = {x: y for x, y in self.group_sizes.items()}
+        self.subpop_from_id = re.compile("^(.*?)[\d]+$")
         self.vicinity = self.parse_vicinity(vicinity)
 
         idx = 0
 
         for ps in popsize:
-            if '=' in ps:
+            if "=" in ps:
                 # this is named population size
-                name, sz = ps.split('=', 1)
+                name, sz = ps.split("=", 1)
             else:
-                name = ''
+                name = ""
                 sz = ps
             try:
                 sz = int(sz)
             except Exception:
                 raise ValueError(
-                    f'Named population size should be name=int: {ps} provided')
+                    f"Named population size should be name=int: {ps} provided"
+                )
 
-            self.add([
-                Individual(
-                    f'{name}_{idx}' if name else str(idx),
-                        susceptibility=getattr(model.params,
-                                           f'susceptibility_mean',
-                                           1) * getattr(model.params,
-                                           f'susceptibility_multiplier_{name}',
-                                           1),
-                    model=model,
-                    logger=logger) for idx in range(idx, idx + sz)
-            ], subpop=name)
+            self.add(
+                [
+                    Individual(
+                        f"{name}_{idx}" if name else str(idx),
+                        susceptibility=getattr(model.params, f"susceptibility_mean", 1)
+                        * getattr(model.params, f"susceptibility_multiplier_{name}", 1),
+                        model=model,
+                        logger=logger,
+                    )
+                    for idx in range(idx, idx + sz)
+                ],
+                subpop=name,
+            )
 
     def parse_vicinity(self, params):
         if not params:
@@ -375,36 +472,50 @@ class Population(object):
 
         res = {}
         for param in params:
-            matched = re.match('^(.*)-(.*)=(\d+)$', param)
+            matched = re.match("^(.*)-(.*)=(\d+)$", param)
             if matched:
                 infector_sp = matched.group(1)
                 infectee_sp = matched.group(2)
                 neighbor_size = int(matched.group(3))
             else:
-                matched = re.match('^(.*)=(\d+)$', param)
+                matched = re.match("^(.*)=(\d+)$", param)
                 if matched:
-                    infector_sp = ''
+                    infector_sp = ""
                     infectee_sp = matched.group(1)
                     neighbor_size = int(matched.group(2))
                 if not matched:
-                    raise ValueError(f'Vicinity should be specified as "INFECTOR_SO-INFECTEE_SP=SIZE": {param} specified')
+                    raise ValueError(
+                        f'Vicinity should be specified as "INFECTOR_SO-INFECTEE_SP=SIZE": {param} specified'
+                    )
 
-            if infector_sp == '':
-                infector_sps = ['']
-            elif infector_sp.startswith('!'):
-                infector_sps = [x for x in self.group_sizes.keys() if not fnmatch(x, infector_sp[1:])]
+            if infector_sp == "":
+                infector_sps = [""]
+            elif infector_sp.startswith("!"):
+                infector_sps = [
+                    x
+                    for x in self.group_sizes.keys()
+                    if not fnmatch(x, infector_sp[1:])
+                ]
             else:
-                infector_sps = [x for x in self.group_sizes.keys() if fnmatch(x, infector_sp)]
+                infector_sps = [
+                    x for x in self.group_sizes.keys() if fnmatch(x, infector_sp)
+                ]
 
-            if infectee_sp.startswith('!'):
-                infectee_sps = [x for x in self.group_sizes.keys() if not fnmatch(x, infectee_sp[1:])]
+            if infectee_sp.startswith("!"):
+                infectee_sps = [
+                    x
+                    for x in self.group_sizes.keys()
+                    if not fnmatch(x, infectee_sp[1:])
+                ]
             else:
-                infectee_sps = [x for x in self.group_sizes.keys() if fnmatch(x, infectee_sp)]
+                infectee_sps = [
+                    x for x in self.group_sizes.keys() if fnmatch(x, infectee_sp)
+                ]
 
-            if infector_sp != '' and not infector_sps:
-                raise ValueError(f'Unrecognized group {infector_sp}')
+            if infector_sp != "" and not infector_sps:
+                raise ValueError(f"Unrecognized group {infector_sp}")
             if not infectee_sps:
-                raise ValueError(f'Unrecognized group {infectee_sp}')
+                raise ValueError(f"Unrecognized group {infectee_sp}")
 
             for infector_sp in infector_sps:
                 for infectee_sp in infectee_sps:
@@ -418,7 +529,7 @@ class Population(object):
         sz = len(self.individuals)
         self.individuals.update({x.id: x for x in items})
         if sz + len(items) != len(self.individuals):
-            raise ValueError(f'One or more IDs are already in the population.')
+            raise ValueError(f"One or more IDs are already in the population.")
         self.group_sizes[subpop] += len(items)
         self.max_ids[subpop] += len(items)
 
@@ -449,15 +560,24 @@ class Population(object):
         # select one non-quarantined indivudal to infect
         #
         if infector is not None and infector not in self.individuals:
-            raise RuntimeError(f'Can not select infectee if since infector {infector} no longer exists.')
+            raise RuntimeError(
+                f"Can not select infectee if since infector {infector} no longer exists."
+            )
 
         # if not cicinity is defines, or
         # if infection is from community and '' not in vicinity, or
         # with infector but self.vicinity does not have this case.
-        if not self.vicinity or (infector is None and '' not in self.vicinity) or \
-            (infector is not None and self.individuals[infector].group not in self.vicinity):
+        if (
+            not self.vicinity
+            or (infector is None and "" not in self.vicinity)
+            or (
+                infector is not None
+                and self.individuals[infector].group not in self.vicinity
+            )
+        ):
             ids = [
-                id for id, ind in self.individuals.items()
+                id
+                for id, ind in self.individuals.items()
                 if infector != ind.id and not ind.quarantined
             ]
         else:
@@ -467,7 +587,7 @@ class Population(object):
             if infector in self.individuals:
                 freq = copy.deepcopy(self.vicinity[self.individuals[infector].group])
             else:
-                freq = copy.deepcopy(self.vicinity[''])
+                freq = copy.deepcopy(self.vicinity[""])
 
             for grp in self.group_sizes.keys():
                 if grp not in freq:
@@ -475,12 +595,14 @@ class Population(object):
                     # now, we know the number of qualified invidiauls from each
                     # group, but we still have excluded and quarantined....
             total = sum(freq.values())
-            freq = {x:y/total for x,y in freq.items()}
+            freq = {x: y / total for x, y in freq.items()}
             # first determine which group ...
             grp = choice(groups, 1, p=[freq[x] for x in groups])
 
             # then select a random individual from the group.
-            ids = [id for id, ind in self.individuals.items()
+            ids = [
+                id
+                for id, ind in self.individuals.items()
                 if ind.group == grp and infector != ind.id and not ind.quarantined
             ]
 
