@@ -145,7 +145,7 @@ def test_get_normal_symptomatic_transmission_probability(default_model):
     N = 10000
     r = []
     for i in range(N):
-        x_grid, prob = default_model.get_symptomatic_transmission_probability(
+        x_grid, prob, _ = default_model.get_symptomatic_transmission_probability(
             incu, R0)
         infected = np.random.binomial(1, prob, len(x_grid))
         r.append(sum(infected))
@@ -160,7 +160,7 @@ def test_get_normal_asymptomatic_transmission_probability(default_model):
     N = 10000
     r = []
     for i in range(N):
-        x_grid, prob = default_model.get_asymptomatic_transmissibility_probability(
+        x_grid, prob, _ = default_model.get_asymptomatic_transmissibility_probability(
             R0)
         infected = np.random.binomial(1, prob, len(x_grid))
         r.append(sum(infected))
@@ -181,7 +181,7 @@ def test_get_piecewise_symptomatic_transmission_probability(default_model):
     N = 10000
     r = []
     for i in range(N):
-        x_grid, prob = default_model.get_symptomatic_transmission_probability(
+        x_grid, prob, _ = default_model.get_symptomatic_transmission_probability(
             incu, R0)
         infected = np.random.binomial(1, prob, len(x_grid))
         r.append(sum(infected))
@@ -200,7 +200,7 @@ def test_get_piecewise_asymptomatic_transmission_probability(default_model):
     N = 10000
     r = []
     for i in range(N):
-        x_grid, prob = default_model.get_asymptomatic_transmissibility_probability(
+        x_grid, prob, _ = default_model.get_asymptomatic_transmissibility_probability(
             R0)
         infected = np.random.binomial(1, prob, len(x_grid))
         r.append(sum(infected))
